
template <int order, typename coef, int step>
    // The order must be positive, and
    // must have a factor that is not a power of two.
    // BAD values are 1, 2, 4, 8, 16, 32, et cetera.
    // All other positive integers are GOOD.
class cos_int_dom {
    static_assert (step >= 1, "cos_int_dom ... step < 1");
    static_assert (step <= 2, "cos_int_dom ... step > 2");

    static constexpr int good_order (int p) {
        return (p < 3) ? false
             : (p & 1) ? true
             : good_order (p / 2);
    }

    static_assert (good_order (order), "cos_int_dom ... bad order");
    offset_array<order,coef,step> guts;

    // begin division support
        using OAC = offset_array<order,coef,step>;
        using OAOAC = offset_array<order,OAC,step>;
        static void divide_row (
            OAC  & numer_row,
            coef & numer_sca,
            coef const & denom);
        static void reduce_row (
            OAC  & row,
            coef & sca);
        static void sub_row (
            OAC  const & subt_row,
            OAC        & minu_row,
            coef const & subt_x,
            coef       & minu_y,
            int  const   col);
        static void view_system (
            OAOAC const & sys,
            OAC   const & sca);
        static void non_integral (
            OAOAC const & sys,
            OAC   const & sca);
    // end division support
public:
    static constexpr int get_order {order};

    cos_int_dom () = delete;
    cos_int_dom (cos_int_dom const &) = default;
    cos_int_dom (cos_int_dom &&) = default;
    cos_int_dom & operator= (cos_int_dom const &) = default;
    cos_int_dom & operator= (cos_int_dom &&) = default;
    ~cos_int_dom () = default;

    // ctor_1
    // use parentheses () to call this. it creates a
    // cos_int_dom whose real-number value equals v.
    explicit cos_int_dom (coef const & v) :
        guts
    {} {
        int p {order};
        int i {2}; int j {4}; int k {4};

        while (p >= 3) {
            if (p & 1) {
                for (int q {i}; q < order; q += k)
                    guts.at(q) = +v;
                for (int q {j}; q < order; q += k)
                    guts.at(q) = -v;
                return;
            }
            p /= 2;
            i *= 2; j *= 2; k *= 2;
        }
        throw fail {"cos_int_dom::ctor_1 ... bad order"};
    }

    // ctor_2
    // use braces {} to enclose an initializer list
    explicit cos_int_dom (initializer_list<coef> ili) :
        guts {}
    {
        if (step == 1) {
            if (ili.size() != ((order - 1) / step))
                throw fail {"cos_int_dom::ctor_2 ... wrong number of initializers"};
            for (int n {1}; n < order; n += 1)
                guts.at(n) = ili.begin()[n - 1];
            return;
        }

        if (order % 2) {
            if (ili.size() != ((order - 1) / step))
                throw fail {"cos_int_dom::ctor_2 ... wrong number of initializers"};
        } else {
            if (ili.size() != ((order - 2) / step))
                throw fail {"cos_int_dom::ctor_2 ... wrong number of initializers"};
        }
        for (int n {2}; n < order; n += 2)
            guts.at(n) = ili.begin()[n / 2 - 1];
    }

    // ctor_3
    // convert lower order to higher
    template <int src_order>
    explicit cos_int_dom (cos_int_dom<src_order,coef,step> const & o) :
        guts {}
    {
        static_assert ((order % src_order) == 0,
            "cos_int_dom::ctor_3 ... order not a multiple"
        );
        constexpr int ratio {order / src_order};
        for (int n {step}; n < src_order; n += step)
            guts.at(n * ratio) = o.at(n);
    }

    // ctor_4
    // convert step 2 to step 1
    template <int src_step>
    explicit cos_int_dom (
        cos_int_dom<order,
        typename std::enable_if <((src_step == 2) && (step == 1)),coef>::type
        ,src_step
    > const & o) :
        guts {}
    {
        static_assert (src_step == 2, "cos_int_dom::ctor_4 ... src_step != 2");
        static_assert (step == 1, "cos_int_dom::ctor_4 ... step != 1");

        for (int n {src_step}; n < order; n += src_step)
            guts.at(n) = o.at(n);
    }

    static cos_int_dom single (int const p) {
        cos_int_dom r (0);
        r.at(p) = coef {1};
        return r;
    }

    ///////////////

    cos_int_dom operator+ () const {
        cos_int_dom r {*this};
        for (int n {step}; n < order; n += step)
            r.guts.at(n) = +r.guts.at(n);
        return r;
    }

    cos_int_dom operator- () const {
        cos_int_dom r {*this};
        for (int n {step}; n < order; n += step)
            r.guts.at(n) = -r.guts.at(n);
        return r;
    }

    ///////////////

    cos_int_dom operator+ (cos_int_dom const & o) const {
        cos_int_dom r {*this};
        for (int n {step}; n < order; n += step)
            r.at(n) += o.guts.at(n);
        return r;
    }

    cos_int_dom operator- (cos_int_dom const & o) const {
        cos_int_dom r {*this};
        for (int n {step}; n < order; n += step)
            r.at(n) -= o.guts.at(n);
        return r;
    }

	cos_int_dom operator* (cos_int_dom const & o) const {
		cos_int_dom r (0);

        for (int p {step}; p < order; p += step) {
            for (int q {step}; q < order; q += step) {
                coef const p_q {guts.at(p) * o.guts.at(q)};
                {
                    int const sum {p + q};
                    if (sum < order)
                        r.guts.at(sum) += p_q;
                    else if (sum > order)
                        r.guts.at(2 * order - sum) -= p_q;
                } {
                    int const dif {p - q};
                    if (dif > 0)
                        r.guts.at(+ dif) += p_q;
                    else if (dif < 0)
                        r.guts.at(- dif) += p_q;
                    else
                        r += cos_int_dom (p_q * 2);
                }
            }
        }

        return r;
	}

	cos_int_dom operator/ (cos_int_dom const &) const;

    cos_int_dom operator* (coef const & o) const {
        cos_int_dom r {*this};
        for (int n {step}; n < order; n += step)
            r.at(n) *= o;
        return r;
    }

    ///////////////

    cos_int_dom & operator+= (cos_int_dom const & o)
        { return *this = *this + o; }

    cos_int_dom & operator-= (cos_int_dom const & o)
        { return *this = *this - o; }

    cos_int_dom & operator*= (cos_int_dom const & o)
        { return *this = *this * o; }

    cos_int_dom & operator/= (cos_int_dom const & o)
        { return *this = *this / o; }

    cos_int_dom & operator*= (coef const & o)
        { return *this = *this * o; }

    ///////////////

    coef & at (int const n) { return guts.at(n); }

    coef at (int const n) const { return guts.at(n); }

    offset_array<order,coef,step> const & at () const { return guts; }

    static double const & two_cos (int const n) {
        static offset_array<order,double,step> table;
        static bool ready {false};

        if (! ready) {
            static double const slice {acos (-1.0) / (2 * order)};
            for (int p {step}; p < order; p += step)
                table.at(p) = 2.0 * cos (p * slice);
            ready = true;
        }

        return table.at(n);
    }

    explicit operator double () const {
        double sum {0.0};
        for (int n {step}; n < order; n += step)
            sum += guts.at(n) * two_cos(n);
        return sum;
    }

    void view (ostream & o) const {
        if (step == 2) o << "<<"; else o << "<";

        for (int n {step}; n < order; n += step) {
            coef const gn {guts.at(n)};
            if (gn == 0) o << noshowpos; else o << showpos;
            o << " " << setw (4) << gn;
        }
        {
            o << noshowpos;
            o << "  [" << order << "]  ";
            double const rn {operator double ()};
            if (rn == 0.0) o << noshowpos; else o << showpos;
            o << setw (dbl_wid) << rn;
        }
        o << noshowpos;
        if (step == 2) o << " >>"; else o << " >";
    }

    static void html (
        ostream &, cos_int_dom const &, bool = false
    );

    static void mul_table ();

    ///////////////

    // possible future projects:
    // -- cross product generalized to n factors in n+1 dimensions
    // -- scalar triple product generalized to n components
};

template <int order, typename coef, int step>
cos_int_dom<order,coef,2> dot (
    cos_int_dom<order,coef,step> const & a,
    cos_int_dom<order,coef,step> const & b
) {
    static constexpr int semiorder {order / 2};
    static cos_int_dom<order,coef,2> const deuce (2);
    cos_int_dom<order,coef,2> r (0);

    static constexpr bool odd {(order % 2) == 1};

    int p {step};

    for ( ; p < semiorder; p += step) {
        coef const ab {a.at(p) * b.at(p)};
        r += deuce * ab;
        r.at (p * 2) += ab;
    }

    for ( ; p <= semiorder; p += step) {
        coef const ab {a.at(p) * b.at(p)};
        r += deuce * ab;
        if (odd) r.at (p * 2) += ab;
    }

    for ( ; p < order; p += step) {
        coef const ab {a.at(p) * b.at(p)};
        r += deuce * ab;
        r.at ((order - p) * 2) -= ab;
    }

    return r;
}

template <int order, typename coef, int step>
bool isomeric (
    cos_int_dom<order,coef,step> const & a,
    cos_int_dom<order,coef,step> const & b
) {
    return isomeric (a.at(), b.at());
}

template <int order, typename coef, int step>
bool anisomeric (
    cos_int_dom<order,coef,step> const & a,
    cos_int_dom<order,coef,step> const & b
) {
    return anisomeric (a.at(), b.at());
}

template <int order, typename coef, int step>
ostream & operator<< (ostream & o, cos_int_dom<order,coef,step> const & t) {
    t.view (o); return o;
}
